/* xn-cal.c:  Pilot calendar conduit
 *
 * Rewritten to handle installing to Palm and reading from it, like
 * xn-addresses.c
 *   Michael J. Hammel, Jan 2004 (mjhammel@graphics-muse.org)
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include "pi-source.h"
#include "pi-socket.h"
#include "pi-dlp.h"
#include "pi-datebook.h"
#include "pi-xn.h"

extern char* optarg;
extern int optind;

char	magicmark[0xffff];
char	*progname;


/*========================================================================
 *	Name:			Help
 *	Prototype:	void Help()
 *					
 *	Description:
 * Print out a help message.
 *
 *	Notes:
 *========================================================================*/
void 
Help(void)
{
  fprintf(stderr,
    "usage:%s [-q] [-p port] [-d category] -a|-r|-w file\n",
    progname);
  fprintf(stderr, "  -q          = do not prompt for HotSync button press\n");
  fprintf(stderr, "  -p port     = use device file <port> to communicate with Pilot\n");
  fprintf(stderr, "  -d category = DELETE old Pilot records in <category>\n");
  fprintf(stderr, "  -r file     = replace existing Calendar db with records read from <file>\n");
  fprintf(stderr, "  -w file     = get records from Pilot and write them to <file>\n");
  exit(2);
}

/*========================================================================
 *	Name:			match_category
 *	Prototype:	void match_category(char *category, struct TodoAppInfo *aai)
 *					
 *	Description:
 * Verify that the category provided is valid based on the categories listed in
 * the tai structure.
 *
 * Returns
 * The index id for the category if it exists.
 * -1 if it doesn't.
 *
 *	Notes:
 *========================================================================*/
int 
match_category(
	char									*buf, 
	struct AppointmentAppInfo		*mai
)
{
	int i;
	for (i=0;i<16;i++)
	if (strcasecmp(buf, mai->category.name[i])==0)
		return i;

	/* Wasn't found - return 0, which equates to the "Unfiled" category. */
	return 0;
}

/*========================================================================
 *	Name:			write_categories
 *	Prototype:	void write_categories(FILE *out, struct AppointmentAppInfo *mai)
 *					
 *	Description:
 * Writes the list of available categories to file.
 *
 * Input arguments:
 * struct	AppointmentAppInfo *mai	Pilot Calendar db's category info, etc.
 * 
 * Output arguments:
 *	FILE		*out			file descriptor to write records to local file
 *
 *	Notes:
 *========================================================================*/
int 
write_categories(
	FILE								*out, 
	struct AppointmentAppInfo	*mai
)
{
	int	i;

	for (i=0;i<16;i++) 
	{
		if ( strlen(mai->category.name[i]) > 0 )
		{
			fprintf(out, "%s\n", mai->category.name[i]);
			printf("Category: %s\n", mai->category.name[i]);
		}
	}

	return 0;
}


/*========================================================================
 *	Name:			read_file
 *	Prototype:	void read_file(FILE *in, int sd, int db, struct AppointmentAppInfo *mai, int mode)
 *					
 *	Description:
 * Read in entries from a text file, formatted the same way we write to
 * text files, and upload it to the pilot.
 *
 * Input arguments:
 *	FILE		*out			file descriptor to reads records from
 *	int		sd				file descriptor used to communicate with pilot 
 *	int		db				database handle information for Pilot Appointment db
 * struct AppointmentAppInfo *mai		Pilot Calendar db's category info, etc.
 *
 * Output arguments:
 *
 * Returns
 * 0 always
 *
 *	Notes:
 *========================================================================*/
int
read_file(
	FILE								*fd,
	int								sd,
	int								db,
	struct AppointmentAppInfo	*mai,
	int								mode
)
{
	int						filelen;
	char						*file_text;
	char						*name, *value;
	int						cPtr;
	int						endPtr;
	int						count;
	unsigned long			recordid = 0;
	unsigned long			new_recordid = 0;
	int						appt_size;
	unsigned char			appt_buf[0xffff];
	int						attributes = 0;
	int						category = 0;
	int						id;
	char						*idchr;
	char						*chr;
	int						j;
	struct Appointment	appt;

	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_text = (char*)malloc(filelen+1);
	if (file_text == NULL) 
	{
		perror("malloc()");
		exit(1);
	}
   fread(file_text, filelen, 1, fd);
	fclose(fd);

	/*
	 * Initialize pointers, counters used to parse input text.
	 */
	cPtr = (int)file_text;
	count = 0;

	/*
	 * Move past the Magic Mark
	 */
	name=strtok((char *)cPtr, "\n");
	cPtr = cPtr + strlen(name) + 1;

	/*
	 * Skip the category fields from the file first.
	 * When we're done, move past the marker that separates the category
	 * list from the records.
	 */
	while( strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 )
	{
		name=strtok((char *)cPtr, "\n");
		cPtr = cPtr + strlen(name) + 1;
	}

	/* Initialize the structure we'll use for each record. */
	memset(&appt,0,sizeof(appt));

	/*
	 * Grab each records data.
	 * 
	 * Record format:
	 * RecordID
	 * Attributes
	 * CategoryID
	 * int Event
	 * struct tm EventStart
	 * struct tm EventEnd
	 * int Alarm
	 * int Advance
	 * int AdvanceUnits
	 * int RepeatType
	 * int RepeatForever
	 * struct tm RepeatEnd
	 * int RepeatFrequency
	 * int RepeatDay
	 * int RepeatDays[7] (int)
	 * int RepeatWeekstart
	 * int Exceptions
	 * struct tm Exception (multiple)
	 * char *Description
	 * char *Note
	 */
	while( cPtr < (int)((int)file_text + filelen) )
	{

		/* Move past the record prefix */
		cPtr = cPtr + strlen(PREFIX);

		/*
		 * Get the current name, value pair.
		 */
		name=strtok((char *)cPtr, ":");
		cPtr = cPtr + strlen(name) + 2;
		value = (char *)cPtr;
	
		while ( 
			(strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 ) && 
			((int)cPtr < (int)((int)file_text + filelen) )
			)
			{
				cPtr++;
			}

		/* 
		 * Remove trailing newline.
		 */
		endPtr = cPtr - 1;

		/*
		 * Null terminate the string 
		 */
		strncpy((char *)endPtr, "\0", 1);


		/* Grab the record id and convert to an integer */
		if ( strncmp(name, "RecordID", strlen("RecordID")) == 0 )
		{
			if ( mode == 1 )
				recordid = (unsigned long)atoi(value);
			else
				recordid = 0;
		}
		else

		/* Grab the record attributes and convert to an integer */
		if ( strncmp(name, "Attributes", strlen("Attributes")) == 0 )
		{
			if ( mode == 1 )
				attributes = atoi(value);
			else
				attributes = 0;
		}
		else

		/* Grab the category name and convert it to an category id */
		if ( strncmp(name, "Category", strlen("Category")) == 0 )
		{
			category = match_category(value,mai);
		}
		else

		/* Grab the event type (ie is it timeless?). */
		if ( strncmp(name, "Event", strlen("Event")) == 0 )
		{
			appt.event = atoi(value);
		}
		else

		/* Grab the Begin date */
		if ( strncmp(name, "Begin", strlen("Begin")) == 0 )
		{
			strptime(value, "%a %b %d %T %Y", &appt.begin);
		}
		else

		/* Grab the End date */
		if ( strncmp(name, "End", strlen("End")) == 0 )
		{
			strptime(value, "%a %b %d %T %Y", &appt.end);
		}
		else

		/* Grab the Alarm */
		if ( strncmp(name, "Alarm", strlen("Alarm")) == 0 )
		{
			appt.alarm = atoi(value);
		}
		else

		/* Grab the Advance */
		if ( strncmp(name, "Advance", strlen("Advance")) == 0 )
		{
			appt.advance = atoi(value);
		}
		else

		/* Grab the AdvanceUnits */
		if ( strncmp(name, "AdvanceUnits", strlen("AdvanceUnits")) == 0 )
		{
			appt.advanceUnits = atoi(value);
		}
		else

		/* Grab the RepeatType */
		if ( strncmp(name, "RepeatType", strlen("RepeatType")) == 0 )
		{
			appt.repeatType = atoi(value);
		}
		else

		/* Grab the RepeatForever */
		if ( strncmp(name, "RepeatForever", strlen("RepeatForever")) == 0 )
		{
			appt.repeatForever = atoi(value);
		}
		else

		/* Grab the RepeatEnd date */
		if ( strncmp(name, "RepeatEnd", strlen("RepeatEnd")) == 0 )
		{
			if ( appt.repeatType == 0 ) 
			{
				bzero(&appt.repeatEnd, sizeof(struct tm));
			}
			else
			{
				strptime(value, "%a %b %d %T %Y", &appt.repeatEnd);
			}
		}
		else

		/* Grab the RepeatFrequency */
		if ( strncmp(name, "RepeatFrequency", strlen("RepeatFrequency")) == 0 )
		{
			appt.repeatFrequency = atoi(value);
		}
		else

		/* Grab the RepeatDay */
		if ( strncmp(name, "RepeatDay", strlen("RepeatDay")) == 0 )
		{
			appt.repeatDay = atoi(value);
		}
		else

		/* Grab the RepeateWeekStart */
		if ( strncmp(name, "RepeastWeekstart", strlen("RepeatWeekstart")) == 0 )
		{
			appt.repeatWeekstart = atoi(value);
		}
		else

		/* Grab the Exceptions */
		if ( strncmp(name, "Exceptions", strlen("Exceptions")) == 0 )
		{
			appt.exceptions = atoi(value);
			appt.exception = (struct tm *)malloc(sizeof(struct tm)*appt.exceptions);
		}
		else

		/* Grab the multiple Exceptions */
		if ( strncmp(name, "ExceptionIDS", strlen("ExceptionIDS")) == 0 )
		{
			id = atoi((char *)(name+12));
			strptime(value, "%a %b %d %T %Y", &appt.exception[id]);
		}
		else


		/* Grab the RepeatDates */
		if ( strncmp(name, "RepeatDates", strlen("RepeatDates")) == 0 )
		{
			idchr = (char *)malloc(strlen(value)+1);
			strcpy(idchr, value);
			bzero((char *)(idchr+strlen(idchr)-1), 1);
			id = atoi(strtok(idchr, ":"));
			for (j=0; j<7; j++)
			{
				appt.repeatDays[j] = id;
				chr = strtok(NULL, ":");
				if ( chr ) id = atoi(chr);
			}
			free(idchr);
		}
		else

		/* Grab the Description */
		if ( strncmp(name, "Description", strlen("Description")) == 0 )
		{
			appt.description = (char *)malloc(strlen(value)+1);
			sprintf(appt.description, "%s", value);
		}
		else

		/* Grab the Note */
		if ( strncmp(name, "Note", strlen("Note")) == 0 )
		{
			appt.note = (char *)malloc(strlen(value)+1);
			sprintf(appt.note, "%s", value);
		}
		else

		if ( strncmp(name, "RecordEND", strlen("RecordEND")) == 0 )
		{
			appt_size = pack_Appointment(&appt, appt_buf, sizeof(appt_buf));
			new_recordid = 0;
			dlp_WriteRecord(sd, (unsigned char)db, attributes, recordid, category,
				appt_buf, appt_size, &new_recordid);

			/* Clean up the appt structure */
			if (appt.exception) free(appt.exception);
			if (appt.description) free(appt.description);
			if (appt.note) free(appt.note);
			memset(&appt,0,sizeof(appt));
		}
	}

	return(0);
}

/*========================================================================
 *	Name:			write_file
 *	Prototype:	void write_file(FILE *out, int sd, int db, struct Appointment *mai) 
 *					
 *	Description:
 * Reads Appointment records from the pilot and writes them to a file.
 *
 * Input arguments:
 *	int		sd				file descriptor used to communicate with pilot 
 *	int		db				database handle information for Pilot Appointment db
 * struct Appointment *mai		Pilot Appointment db's category info, etc.
 * 
 * Output arguments:
 *	FILE		*out			file descriptor to write records to local file
 *
 *	Notes:
 *========================================================================*/
int 
write_file(
	FILE						*out,
	int						sd,
	int						db,
	struct AppointmentAppInfo	*mai
)
{
	int				attr;
	int				category;
	int				len;
	int				i,j;
	struct Appointment		appt;
	unsigned char	buffer[0xffff];
	recordid_t		recordid;

	/*
	 * First, clean up the database. This will actually remove the
	 * archived and deleted files on the Pilot so we don't download them.
	 */
	dlp_CleanUpDatabase(sd, db);

	for (i=0;;i++) 
	{
		len = dlp_ReadRecordByIndex(sd, db, i, buffer, 
					&recordid, 0, &attr, &category);

		if(len<0)
			break;
    	
		/* Skip deleted records */
		if((attr & dlpRecAttrDeleted) || (attr & dlpRecAttrArchived))
			continue;
    
		unpack_Appointment(&appt, buffer, len);

		/* Write the record to the file. */
		fprintf(out, "%sRecordID: %d\n", PREFIX, (int)recordid);
		fprintf(out, "%sAttributes: %d\n", PREFIX, (int)attr);
		fprintf(out, "%sEvent: %d\n", PREFIX, appt.event);
		fprintf(out, "%sBegin: %s", PREFIX, asctime(&appt.begin));
		fprintf(out, "%sEnd: %s", PREFIX, asctime(&appt.end));
		fprintf(out, "%sAlarm: %d\n", PREFIX, appt.alarm);
		fprintf(out, "%sAdvance: %d\n", PREFIX, appt.advance);
		fprintf(out, "%sAdvanceUnits: %d\n", PREFIX, appt.advanceUnits);
		fprintf(out, "%sRepeatType: %d\n", PREFIX, appt.repeatType);
		fprintf(out, "%sRepeatForever: %d\n", PREFIX, appt.repeatForever);
		if ( appt.repeatType == 0 ) bzero(&appt.repeatEnd, sizeof(struct tm));
		fprintf(out, "%sRepeatEnd: %s", PREFIX, asctime(&appt.repeatEnd));
		fprintf(out, "%sRepeatFrequency: %d\n", PREFIX, appt.repeatFrequency);
		fprintf(out, "%sRepeatDay: %d\n", PREFIX, appt.repeatDay);
		fprintf(out, "%sRepeatDates: ", PREFIX);
		for (j=0; j<7; j++)
			fprintf(out, "%d:", appt.repeatDays[j]);
		fprintf(out, "\n");
		fprintf(out, "%sRepeatWeekstart: %d\n", PREFIX, appt.repeatWeekstart);
		fprintf(out, "%sExceptions: %d\n", PREFIX, appt.exceptions);
		for (j=0; j<appt.exceptions; j++)
			fprintf(out, "%sExceptionIDS%d: %s", 
				PREFIX, j, asctime(&appt.exception[j]));
		if ( appt.description )
			fprintf(out, "%sDescription: %s\n", PREFIX, appt.description);
		else
			fprintf(out, "%sDescription: \n", PREFIX);
		if ( appt.note )
			fprintf(out, "%sNote: %s\n", PREFIX, appt.note);
		else
			fprintf(out, "%sNote: \n", PREFIX);
		fprintf(out, "%sRecordEND:\n", PREFIX);
	}

	return(0);
}


/*========================================================================
 *	Name:			main
 *					
 *	Description:
 * Let's go!
 *========================================================================*/
int 
main(
	int argc, 
	char *argv[]
)
{
	struct pi_sockaddr	addr;
	int						db;
	int						sd;
	int						l;
	FILE						*fd;
	struct PilotUser		U;
	int						ret;
	char						buf[0xffff];
	struct AppointmentAppInfo	mai;
	char						*progname, *category_name;
	int						quiet = 0;
	int						replace_category;

	int						mode = 0;
	int						c;
	char						*deletecategory = 0;

	category_name = NULL;
	replace_category = 0;

	progname = argv[0];

	sprintf(magicmark, MAGICMARK, "datebook");

	/*
	 * Check for PILOTPORT environment variable if user doesn't specify the
	 * port to use on the command line.
	 */
	if (getenv("PILOTPORT"))
	{
		strcpy(addr.pi_device,getenv("PILOTPORT"));
	}
	else
	{
		strcpy(addr.pi_device,PILOTPORT);
	}

	/*
	 * If we have less than 3 cmd lin args, print the help message.
	 */
	if (argc < 3)
		Help();


	/*
	 * Parse command line options, if any.
	 */
	while (((c = getopt(argc, argv, "qap:d:rw")) != -1) && (mode == 0))
	{
		switch (c)
		{
		case 'q':
			quiet = 1;
			break;
		case 'p':
			/* optarg is name of port to use instead of $PILOTPORT or /dev/pilot */
			strcpy(addr.pi_device,optarg);
			break;
		case 'd':
			deletecategory = optarg;
			break;
		case 'r':
			mode = 1;
			break;
		case 'w':
			mode = 2;
			break;
		case 'h': case '?':
			Help();
			break;
		}

	}

	/*
	 * -r or -w argument is required.  If not found, print help message.
	 */
	if (mode == 0)
		Help();

	/*
	 * Unless silent run is requested, prompt for pressing the HotSync button on
	 * the Pilot or cradle.
	 */
	if (!quiet)
	{
		fprintf(stderr,
			"Please insert Pilot in cradle on %s and press HotSync button.\n",
			addr.pi_device);
	}

	/*
	 * Get a socket and set up for communicating with the Pilot.
	 */
	if (!(sd = pi_socket(PI_AF_PILOT, PI_SOCK_STREAM, PI_PF_DLP))) 
	{
		perror("pi_socket");
		exit(3);
	}
    
	addr.pi_family = PI_AF_PILOT;
  
	ret = pi_bind(sd, (struct sockaddr*)&addr, sizeof(addr));
	if(ret == -1) 
	{
		perror("pi_bind");
		exit(3);
	}

	ret = pi_listen(sd,1);
	if(ret == -1) 
	{
		perror("pi_listen");
		exit(3);
	}

	sd = pi_accept(sd, 0, 0);
	if(sd == -1) 
	{
		perror("pi_accept");
		exit(3);
	}

	/* Ask the pilot who it is. */
	dlp_ReadUserInfo(sd,&U);
  
	/* Tell user (via Pilot) that we are starting things up */
	dlp_OpenConduit(sd);
  
	/* Open the Appointment Pad's database, store access handle in db */
	if(dlp_OpenDB(sd, 0, 0x80|0x40, "DatebookDB", &db) < 0) 
	{
		puts("Unable to open DatebookDB");
		dlp_AddSyncLogEntry(sd, "Unable to open DatebookDB.\n");
		exit(1);
	}
  
	l = dlp_ReadAppBlock(sd, db, 0, (unsigned char *)buf, 0xffff);
	unpack_AppointmentAppInfo(&mai, (unsigned char *)buf, l);


	/*
	 * If the request is to retrieve records from the Pilot and save to a local
	 * file (ie the -w option was used)...
	 */
	switch (mode)
	{
	case 2:
		/*
		 * Open the output file.
		 */
		fd = fopen(argv[optind],"w");
		if (fd == NULL)
		{
			sprintf(buf, "%s: %s", argv[0], argv[optind]);
			perror(buf);
			exit(1);
		}
		fprintf(fd, "%s\n", magicmark);

		/*
		 * Read from the pilot and saved to a file.
		 */
		write_categories(fd, &mai);
		write_file(fd, sd, db, &mai);
		
		/*
		 * if requested, delete the category and it's entries.  Only do this if
		 * the category is *not* the Unfiled category.
		 */
		if (deletecategory)
			dlp_DeleteCategory(sd, db, match_category(deletecategory,&mai));
		fclose(fd);
		break;

	case 1:

		/*
       * The -r option was specified, so write a text file full of
       * Appointment entries to the Pilot.
		 */

		/* ... open the output file. */
		fd = fopen(argv[optind], "r");
		if (fd == NULL) 
		{
			sprintf(buf, "%s: %s", argv[0], argv[optind]);
			perror(buf);
			exit(1);
		}

		/*
		 * Delete all the old records first.
		 */
		if ( mode == 1 )
			dlp_DeleteRecord(sd, db, 1, 0);

		/*
		 * Process the records retrieved.
		 */
		read_file(fd, sd, db, &mai, mode);

		/*
		 * if requested, delete the category and it's entries.  Only do this if
		 * the category is *not* the Unfiled category.
		 */
		if (deletecategory)
			dlp_DeleteCategory(sd, db, match_category(deletecategory,&mai));

		break;

	}

	/*
	 * Now clean up the database. This will actually remove the
	 * archived and deleted files on the Pilot.
	 */
	dlp_CleanUpDatabase(sd, db);

	/* Close the database */
	dlp_CloseDB(sd, db);

	/* Tell the user who it is, with a different PC id. */
	U.lastSyncPC = 0x00010000;
	U.successfulSyncDate = time(NULL);
	U.lastSyncDate = U.successfulSyncDate;
	dlp_WriteUserInfo(sd,&U);

	dlp_AddSyncLogEntry(sd, "Wrote appointments(s) to Pilot.\n");
  
	/* All of the following code is now unnecessary, but harmless */
  
	dlp_EndOfSync(sd,0);
	pi_close(sd);
	
	return 0;
}
